from abc import ABCMeta, abstractmethod, abstractproperty


class IConnector:
    """
    Abstract class for db-connections
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def exec(self, query_str: str, identifiers=None, literals=None, result='all'):
        """
        :param query_str: request string,
            containing symbols.QRDB_IDENTIFIER for identifiers and symbols.QRDB_LITERAL for literals
        :param identifiers: iterable of identifiers
        :param literals: iterable of literals
        :param result: one of 'all' and 'one' - amount of rows to return from query results
        :return: DBResult instance
        """

    @abstractmethod
    def table_info(self):
        """
        return system info about tables in chosen db in given format:
        {'books':[('id', 'integer'), ('date', 'date'), ...], ...}
        """

    @abstractmethod
    def commit(self):
        """
        commit changes to database
        """


# todo are abstracts needed?
# todo add about success
class DBResult:
    """
    class for db-query data.
    """

    def __init__(self, data, result_type):
        self.result_type = result_type
        self.data = data
        self.used_fields = []

    def set_used_fields(self, used_fields):
        self.used_fields = used_fields

    def get_data(self):
        return self.data

    def get_result_type(self):
        return self.result_type

    def get_used_fields(self):
        return self.used_fields

    def is_one_result(self):
        return self.result_type == 'one'

    def is_all_result(self):
        return self.result_type == 'all'

    def is_no_result(self):
        return self.result_type is None


from abc import ABCMeta, abstractmethod, abstractproperty
from data import *
import operators as op
from collections.abc import Iterable
from symbols import *


def get_field(field_name: str, tables) -> QRField:
    """
    :param field_name: field name
    :param tables: iterable of QRTables
    :return: QRField corresponding to given name or error,
    if name wasn't found or found several suitable fields
    """

    cnt = 0
    field = None
    for t in tables:
        if t.__dict__.get(field_name) is not None:
            if cnt:
                raise Exception('ambiguous attribute name %s' % field_name)
            field = t.__dict__[field_name]
    if field is None:
        raise Exception('attribute %s not found' % field_name)
    return field


def parse_request_args(tables, *args, disable_full_name=False, **kwargs):
    """
    :param tables: list of QRTable - tables used in query
    :param disable_full_name: if False, identifiers will be extended with table name: like 'id'->'table.id'
    :return: 3 lists: identifiers, literals, query parts (conditions)
    """
    identifiers = []
    literals = []
    conditions = []
    for field_name, condition in kwargs.items():
        if not isinstance(condition, op.IQROperator):
            condition = op.Eq(condition)

        condition, lits = condition.condition(disable_full_name)
        field = get_field(field_name, tables)

        if disable_full_name:
            identifiers.extend([field.name])
        else:
            identifiers.extend([field.table_name, field.name])
        literals.extend(lits)
        conditions.append(condition)

    for arg in args:
        # todo else warning of unknown
        if isinstance(arg, str):
            conditions.append(arg)
        elif isinstance(arg, op.Eq):
            if arg.has_both_args():
                condition, _ = arg.condition(disable_full_name)
                ids = arg.arg1, arg.arg2
                identifiers.extend(ids)
                conditions.append(condition)
    return identifiers, literals, conditions


class QueryPartData:
    """structured information for IQueryPart accumulated results"""
    def __init__(self, identifiers, literals, query: str):
        self.identifiers = identifiers
        self.literals = literals
        self.query = query


class IQueryPart:
    """
    Abstract class for db-query parts like group_by, where, limit etc.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def add_data(self, *args, **kwargs) -> QueryPartData:
        """use query part with given arguments; example: obj.where(id=1, name=In(['a','b']))"""

    @abstractmethod
    def get_name(self):
        """get name of query part - the one is used as a method name in QRQuery, so it must be a valid identifier"""

    @abstractmethod
    def get_data(self) -> QueryPartData:
        """get identifiers, literals and query string accumulated in query part (multiple calls for single
        IQueryPart object is possible)"""

    @abstractmethod
    def add_tables(self, tables):
        """add list of tables"""

    @abstractmethod
    def set_tables(self, tables):
        """set list of tables to use"""


class QueryPart(IQueryPart):
    def __init__(self, name: str, command: str = None, tables=None):
        if tables is None:
            tables = []
        if command is None:
            command = name
        self.tables = tables
        self.command = command
        self.identifiers, self.literals = [], []
        self.query = None
        self.name = name

    def add_conditions(self, conditions, joint=' '):
        for cond in conditions:
            if self.query is None:
                self.query = self.command + ' ' + cond
            else:
                self.query += ' ' + joint + ' ' + cond

    def get_data(self):
        return QueryPartData(self.identifiers, self.literals, self.query if self.query else '')

    def get_name(self):
        return self.name

    def add_data(self, qr_joint, *args, **kwargs):
        identifiers, literals, conditions = parse_request_args(self.tables, *args, **kwargs)
        self.identifiers.extend(identifiers)
        self.literals.extend(literals)

        self.add_conditions(conditions, joint=qr_joint)

    def add_tables(self, *tables):
        self.tables += tables

    def set_tables(self, tables):
        self.tables = tables


class QRWhere(QueryPart):
    def __init__(self, tables=None):
        super().__init__('where', tables=tables)

    def add_data(self, *args, **kwargs):
        op = kwargs.get('bool')
        if op is not None:
            kwargs.pop('bool')
        if op not in ['and', 'or']:
            op = 'and'

        super(QueryPart, self).add_data(qr_joint=op, *args, **kwargs)


class QRGroupBy(QueryPart):
    def __init__(self, tables=None):
        super().__init__(name='group_by', command='group by', tables=tables)

    def add_data(self, *args, **kwargs):
        for field in args:
            self.identifiers.append(field.name)
        self.add_conditions([QRDB_IDENTIFIER] * len(args), joint=',')


class QRHaving(QueryPart):
    def __init__(self, tables=None):
        super().__init__('having', tables=tables)

    def add_data(self, *args, **kwargs):
        super().add_data(qr_joint=',', *args, **kwargs)


class QROrderBy(QueryPart):
    def __init__(self, tables=None):
        super().__init__(name='order_by', command='order by', tables=tables)

    def add_data(self, *args, **kwargs):
        sort_type = 'asc'
        if kwargs.get('desc') == True:
            sort_type = 'desc'

        for field in args:
            self.identifiers.append(field.name)

        self.add_conditions([QRDB_IDENTIFIER + ' ' + sort_type] * len(args), joint=',')


class QRLimit(QueryPart):
    def __init__(self, tables=None):
        super().__init__('limit', tables=tables)

    def add_data(self, *args, **kwargs):
        if len(kwargs) > 0 or len(args) != 1:
            raise Exception('Limit condition error: only one integer expected')
        n = args[0]
        if not isinstance(n, int):
            raise Exception('Integer expected as a limit condition param')

        self.literals.append(n)
        self.query = ' limit ' + QRDB_LITERAL + ' '


class QROffset(QueryPart):
    def __init__(self, tables=None):
        super().__init__('offset', tables=tables)

    def add_data(self, *args, **kwargs):
        if len(kwargs) > 0 or len(args) != 1:
            raise Exception('Offset condition error: only one integer expected')
        n = args[0]
        if not isinstance(n, int):
            raise Exception('Offset expected as a limit condition param')

        self.literals.append(n)
        self.query = ' offset ' + QRDB_LITERAL + ' '


class QRJoin(QueryPart):
    def __init__(self, tables=None):
        super().__init__('join', tables=tables)

    def add_data(self, *args, **kwargs):
        if len(kwargs) > 0 or len(args) != 2:
            raise Exception('Join condition error: two conditional params')
        table, cond = args[0], args[1]
        self.add_tables(table)
        # todo accurate with adding - links working for all query-parts!

        if type(cond) == str:
            self.identifiers.append(table.meta['table_name'])
            join_cond = QRDB_IDENTIFIER + ' on %s' % cond

        else:
            # todo different imports -> no recognition if not isinstance(cond, op.Eq):
            #    raise Exception('join: op.Eq instance expected, got %s' % type(cond))
            if not cond.has_both_args():
                raise Exception('join: op.Eq contains only one instance')
            if not isinstance(cond.arg1, QRField) or not isinstance(cond.arg2, QRField):
                raise Exception('join: op.Eq operands must be QRField instances')
            if cond.arg1.table not in self.tables or cond.arg1.table not in self.tables:
                raise Exception('join: wrong attribute\'s relations')

            self.identifiers.extend([table.meta['table_name'],
                                     cond.arg1.table_name, cond.arg1.name,
                                     cond.arg2.table_name, cond.arg2.name])
            join_cond = '%s on %s.%s = %s.%s' % (QRDB_IDENTIFIER, QRDB_IDENTIFIER,
                                                 QRDB_IDENTIFIER, QRDB_IDENTIFIER, QRDB_IDENTIFIER)

        self.add_conditions([join_cond], joint='join')


class QRSet(QueryPart):
    def __init__(self, tables=None):
        super().__init__('set', tables=tables)

    def add_data(self, *args, **kwargs):
        # todo here stuff like x>10 possible... in set - no sense
        identifiers, literals, conditions = parse_request_args(self.tables,
                                                               *args, **kwargs, disable_full_name=True)
        self.identifiers.extend(identifiers)
        self.literals.extend(literals)

        self.add_conditions(conditions, joint=',')


class QRValues(QueryPart):
    def __init__(self, column_cnt, tables=None):
        super().__init__('values', tables=tables)
        self.column_cnt = column_cnt

    def add_data(self, *args, **kwargs):
        literals = []
        value_sets = 0

        for arg in args:
            if not isinstance(arg, Iterable):
                raise Exception('insert values: iterable expected')
            flag = False
            for a in arg:
                if isinstance(a, Iterable) and not isinstance(a, str):
                    if len(a) != self.column_cnt:
                        raise Exception('insert values: wrong column count')
                    literals.extend(a)
                    value_sets += 1
                else:
                    literals.append(a)
                    flag = True
            if flag:
                value_sets += 1

        single = (QRDB_LITERAL + ',') * self.column_cnt
        single = '(' + single[:-1] + ')'
        query = ', '.join([single] * value_sets)

        if self.query is None:
            self.query = query
        else:
            self.query += ', ' + query
        self.literals.extend(literals)


class QRReturning(QueryPart):
    def __init__(self, used_fields, tables=None):
        super().__init__('returning', tables=tables)
        self.column_cnt = len(self.identifiers) - 1
        self.used_fields = used_fields

    def add_data(self, *args, **kwargs):
        identifiers = []
        query = []

        def parse_arg(arg, query):
            if isinstance(arg, QRField):
                query += [QRDB_IDENTIFIER]
                identifiers.extend([arg.name])
                self.used_fields.append(arg.name)
            else:
                if arg == '*':
                    self.used_fields.extend(list(self.tables[0].meta['fields'].keys()))
                else:
                    self.used_fields.append(arg)
                query += [arg]

        for arg in args:
            if isinstance(arg, Iterable) and not isinstance(arg, str):
                for a in arg:
                    parse_arg(a, query)
            else:
                parse_arg(arg, query)


        self.identifiers.extend(identifiers)
        self.add_conditions(query, joint=',')


import inject
from abc import ABCMeta, abstractmethod, abstractproperty
from IConnector import IConnector
from data_formatter import IDataFormatter

from query_parts import *
from symbols import *
import qrlogging


class IQRQuery:
    """
    Abstract class for db-queries like select, update, delete  etc.
    """

    @abstractmethod
    def exec(self, result: str = None):
        """
        execute the formed query
        :param result: one of 'all', 'one' - number of rows to get from query result
        :return: None if results is None, else data in format specified by injected IDataFormatter instance
        WARNING - if result is 'one', no 'limit 1' is added to query - one must specify that yourself
        """

    @abstractmethod
    def all(self):
        """shortcut for exec with result='all'"""

    @abstractmethod
    def one(self):
        """shortcut for exec with result='one'"""

    def get_error(self):
        """return None or str with description of error occured while making query"""


class QRQuery(IQRQuery):
    data_formatter = inject.attr(IDataFormatter)

    def __init__(self, connector: IConnector, table: QRTable = None, query: str = '',
                 identifiers=None, literals=None, used_fields=None, auto_commit=False):
        self.connector = connector
        self.query = query
        if identifiers is None:
            identifiers = []
        if literals is None:
            literals = []
        if used_fields is None:
            used_fields = []
        self.identifiers = identifiers
        self.literals = literals
        self.tables = [table] if table else []
        self.conditions = dict()
        self.auto_commit = auto_commit
        self.cur_order = -1
        self.used_fields = used_fields
        self.query_parts = []
        self.error = None

    def get_error(self):
        return self.error

    def __create_method(self, qp: IQueryPart):
        n = len(self.query_parts)

        def f(*args, **kwargs):
            if self.error is not None:
                return self
            if self.cur_order > n:
                self.error = 'select: wrong operators sequence'
            self.cur_order = n

            try:
                qp.add_data(*args, **kwargs)
            except Exception as e:
                qrlogging.exception(e)
                self.error = str(e)
            return self
        return f

    def _add_query_part(self, qp: IQueryPart):
        qp.set_tables(self.tables)
        method_name = qp.get_name()
        self.__dict__[method_name] = self.__create_method(qp)
        self.query_parts.append(qp)

    @log_error_default()
    def exec(self, result=None):
        if self.error is not None:
            return None

        for qp in self.query_parts:
            data = qp.get_data()
            self.identifiers.extend(data.identifiers)
            self.literals.extend(data.literals)
            self.query += ' ' + data.query

        data = self.connector.exec(self.query, self.identifiers, self.literals, result=result)
        data.set_used_fields(self.used_fields)

        if self.auto_commit:
            self.connector.commit()

        return True if (result is None) else \
            self.data_formatter.format_data(data)

    def all(self):
        return self.exec('all')

    def one(self):
        return self.exec('one')


class QRSelect(QRQuery):
    def __init__(self, connector: IConnector, table: QRTable, *args):
        try:
            identifiers, literals, used_fields = [], [], dict()
            if len(args) == 0:
                own_args = list(table.meta['fields'].values())
            else:
                own_args = args

            fields = ''
            for arg in own_args:
                if isinstance(arg, QRField):
                    fields += '%s.%s,' % (QRDB_IDENTIFIER, QRDB_IDENTIFIER)
                    identifiers.extend([arg.table_name, arg.name])
                    self.__add_used_field(used_fields, {'name': arg.name, 'table': arg.table_name})
                elif isinstance(arg, str):  # todo else warning
                    fields += arg + ','
                    self.__add_used_field(used_fields, {'name': arg})
            fields = fields[:-1]

            used_fields = [k for k, v in used_fields.items() if not v.get('expired')]

            query = 'select ' + fields + ' from ' + QRDB_IDENTIFIER
            table_name = table.meta['table_name']
            identifiers += [table_name]

            super().__init__(connector, table, query=query,
                             identifiers=identifiers, literals=literals, used_fields=used_fields)
        except Exception as e:
            super().__init__(connector, table)
            self.error = str(e)
            qrlogging.warning("Failed to init select-query: %s", e)
        finally:
            self.__configure_query_parts()

    def __configure_query_parts(self):
        # order is important - must correlate with query parts order
        self._add_query_part(QRJoin(self.tables))
        self._add_query_part(QRWhere(self.tables))
        self._add_query_part(QRGroupBy(self.tables))
        self._add_query_part(QRHaving(self.tables))
        self._add_query_part(QROrderBy(self.tables))
        self._add_query_part(QRLimit(self.tables))
        self._add_query_part(QROffset(self.tables))

    def __add_used_field(self, fields, a):
        x = fields.get(a['name'])
        if x is not None:
            if x.get('table') is None or a.get('table') is None:
                raise Exception('two identical return fields set for select query')
            elif x.get('table') == a.get('table'):
                raise Exception('two identical return fields set for select query')
            elif x.get('expired') is None:
                fields[x['table'] + '_' + x['name']] = x.copy()
                fields[x['name']]['expired'] = True
                fields[x['name']]['tables'] = [x['table'], a['table']]
            else:
                if a['table'] in fields[a['name']]['tables']:
                    raise Exception('two identical return fields set for select query')
                fields[x['name']]['tables'].append(a['table'])
            fields[a['table'] + '_' + a['name']] = a
        else:
            fields[a['name']] = a


class QRUpdate(QRQuery):
    def __init__(self, connector: IConnector, table: QRTable, auto_commit=False):
        try:
            identifiers, literals = [], []

            query = 'update ' + QRDB_IDENTIFIER
            table_name = table.meta['table_name']
            identifiers += [table_name]

            super().__init__(connector, table, query=query,
                             identifiers=identifiers, literals=literals, auto_commit=auto_commit)
            self.__configure_query_parts()
        except Exception as e:
            super().__init__(connector, table)
            self.error = str(e)
            qrlogging.warning("Failed to init update-query: %s", e)
        finally:
            self.__configure_query_parts()

    def __configure_query_parts(self):
        self._add_query_part(QRSet(self.tables))
        self._add_query_part(QRWhere(self.tables))


class QRDelete(QRQuery):
    def __init__(self, connector: IConnector, table: QRTable, auto_commit=False):
        try:
            identifiers, literals = [], []

            query = 'delete from ' + QRDB_IDENTIFIER
            table_name = table.meta['table_name']
            identifiers += [table_name]

            super().__init__(connector, table, query=query,
                             identifiers=identifiers, literals=literals, auto_commit=auto_commit)
            self.__configure_query_parts()
        except Exception as e:
            super().__init__(connector, table)
            self.error = str(e)
            qrlogging.warning("Failed to init select-query: %s", e)
        finally:
            self.__configure_query_parts()

    def __configure_query_parts(self):
        self._add_query_part(QRWhere(self.tables))


@log_class(log_error_default_self)
class QRInsert(QRQuery):
    def __init__(self, connector: IConnector, table: QRTable, *args, auto_commit=False):
        try:
            identifiers, literals = [], []
            if len(args) == 0:
                fields = ''
            else:
                fields = ''
                for arg in args:
                    if isinstance(arg, QRField):
                        fields += QRDB_IDENTIFIER + ' ,'
                        identifiers.extend([arg.name])
                    else:
                        fields += arg + ','
                fields = fields[:-1]
                fields = '(' + fields + ')'

            query = 'insert into ' + QRDB_IDENTIFIER + ' ' + fields + ' values '
            table_name = table.meta['table_name']
            identifiers = [table_name] + identifiers

            super().__init__(connector, table, query=query,
                             identifiers=identifiers, literals=literals, auto_commit=auto_commit)
            self.__configure_query_parts()
        except Exception as e:
            super().__init__(connector, table)
            self.error = str(e)
            qrlogging.warning("Failed to init select-query: %s", e)
        finally:
            self.__configure_query_parts()

    def __configure_query_parts(self):
        self._add_query_part(QRValues(tables=self.tables, column_cnt=len(self.identifiers) - 1))
        self._add_query_part(QRReturning(self.used_fields, self.tables))


@log_class(log_error_default_self)
class QRExec(QRQuery):
    def __init__(self, connector: IConnector, raw_query, auto_commit=False):
        super().__init__(connector, query=raw_query, auto_commit=auto_commit)

    def config_fields(self, *args):
        """set field names to return"""
        # todo add support for iterable params
        self.used_fields = list(args)
        return self



from abc import ABCMeta, abstractmethod, abstractproperty
from error_handlers import log_error
from IConnector import DBResult


class IDataFormatter:
    """
    Abstract class for db-connections
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def format_data(self, data: DBResult):
        """
        :param query_str: request string, containing {} for identifiers and %s for literals
        :param identifiers: iterable of identifiers
        :param literals: iterable of literals
        :param result: one of 'all' and 'one' - amount of rows to return from query results
        :return: for result='all': [(1, 2, 3), ...] or []
                 for result='one': (1,2,3) or None
        """


class ListDataFormatter(IDataFormatter):
    def format_data(self, data: DBResult):
        return data.get_data()


class DictDataFormatter(IDataFormatter):
    def format_data(self, data: DBResult):
        if data.is_no_result():  # todo check for None data was here long ago... needed?
            return None

        used_fields = data.get_used_fields()
        dt = data.get_data()
        res = []
        n = len(used_fields)
        if data.is_one_result():
            return {used_fields[i]: dt[i] for i in range(n)}
        elif data.is_all_result():
            for d in dt:
                res.append({used_fields[i]: d[i] for i in range(n)})
            return res
        else:
            raise Exception("unexpected data result: neither one nor all")


import DI
import sys
from query import *
import operators

# todo-list throughout the orm-project
# important
# todo tests

# middle
# todo insert values -> add array of dicts support

# future
# todo add support for nested queries
# todo add select(books.id) parsing -> extract table from ONLY datafield
# todo add returning part for delete query (and update?)
# todo add StructDataFormatter
# todo add 'as' syntax: 'select count(*) as cnt


import qrlogging

class DBQueryAggregator:
    def __init__(self, conn: IConnector):
        self.connector = conn

    def exec(self, raw_query):
        return QRExec(self.connector, raw_query)

    def select(self, table: QRTable, *args, **kwargs):
        return QRSelect(self.connector, table, *args)

    def delete(self, table: QRTable, auto_commit=False, *args, **kwargs):
        return QRDelete(self.connector, table, auto_commit)

    def update(self, table: QRTable, auto_commit=False, *args, **kwargs):
        return QRUpdate(self.connector, table, auto_commit)

    def insert(self, table: QRTable, *args, auto_commit=False, **kwargs):
        return QRInsert(self.connector, table, *args, auto_commit=auto_commit)


class DBCreator:
    """
    Class responsible for objects creation
    """

    def __init__(self, conn: IConnector, db: DBQueryAggregator):
        """
        :param conn: active connection for tables to use
        :param db: query aggregator
        """
        self.conn = conn
        self.db = db

    def create_data(self, source=None, in_module=False):
        """creates table data (described in DB.create_data) and returns it"""
        tables = self.conn.table_info()
        t = dict()

        for name, field in tables.items():
            t[name] = QRTable(name, field, self.db)

        self.__dict__.update(t)
        if source:
            if in_module:
                source = sys.modules[source]
            source.__dict__.update(t)
        return t


class DB:
    operators = operators

    def __init__(self, connector_type, *conn_args, format_type=None, **conn_kwargs):
        """
        :param connector_type: now only 'postgres' is supported
        :param conn_args: arg-params like host and port needed to connect to database
        :param format_type: the format data will be returned in; now only 'list' and 'dict' are supported
        :param conn_kwargs: kwarg-params like host and port needed to connect to database
        """
        DI.register(format_type, connector_type, *conn_args, **conn_kwargs)

        self.meta = dict()
        self.meta['connector'] = inject.instance(IConnector)
        self.meta['aggregator'] = DBQueryAggregator(self.meta['connector'])

    def create_logger(self, logger_name='default', app_name='app',
                  level="INFO", file: str = None, file_level="INFO"):
        qrlogging.logger = qrlogging.create_logger(logger_name, app_name, level, file, file_level)

    def create_data(self, source=None, in_module=False):
        """ Create QRTable objects for all tables in given database (use system tables data)
            and stores these objects in this class.
            if 'source' present, attributes with names of found tables will be added to it
            'source' is object with __dict__ field or a module name (with 'in_module' flag up)
        """
        data = DBCreator(self.meta['connector'], self.meta['aggregator']).create_data(source, in_module)
        self.__dict__.update(data)

    def commit(self):
        self.meta['connector'].commit()

    def exec(self, raw_query):
        return self.meta['aggregator'].exec(raw_query)

    def select(self, table: QRTable, *args):
        return self.meta['aggregator'].select(table, *args)

    def delete(self, table: QRTable, auto_commit=False):
        return self.meta['aggregator'].delete(table, auto_commit)

    def update(self, table: QRTable, auto_commit=False):
        return self.meta['aggregator'].update(table, auto_commit)

    def insert(self, table: QRTable, *args, auto_commit=False):
        return self.meta['aggregator'].insert(table, *args, auto_commit=auto_commit)




from error_handlers import *

@log_class(log_error)
@exc_no_db(exceptions=['set_DB'])
class QRTable:
    """
    Object representing a table in database.
    Contains info about table (name, fields as QRField objects)
    and functions for quick access to queries
    """

    def __init__(self, table_name=None, fields=None, DB=None):
        """
        :param fields: in format {'field_name': 'field_type', ...}
        :param DB: DBQueryAggregator instance
        """
        self.meta = dict()
        self.meta['table_name'] = table_name
        self.meta['fields'] = {}
        self._DB = DB

        if fields is None: return
        for name, value_type in fields:
            f = QRField(name, value_type, self)
            self.meta['fields'][name] = f
            self.__dict__[name] = f

    def __str__(self):
        if self.meta['table_name'] is None:
            return '<Empty QRTable>'
        return '<QRTable ' + self.meta['table_name'] + '>'

    def set_DB(self, DB):
        self._DB = DB

    def select(self, *args, **kwargs):
        return self._DB.select(self, *args, **kwargs)

    def update(self, *args, **kwargs):
        return self._DB.update(self, *args, **kwargs)

    def insert(self, *args, **kwargs):
        return self._DB.insert(self, *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._DB.delete(self, *args, **kwargs)


class QRField:
    """
    Object representing a table's field in database.
    """

    def __init__(self, name, value_type, table: QRTable):
        self.name = name
        self.type = value_type
        self.table_name = table.meta['table_name']
        self.table = table

    def __str__(self):
        if self.name is None:
            return '<Empty QRField>'
        return '<QRField ' + self.name + ' of ' + self.table_name + '>'



import psycopg2
import psycopg2.sql as sql

from IConnector import IConnector, DBResult
from error_handlers import log_error, retry_log_error
from threading import Lock
import symbols
import qrlogging

# IConnector realization for Postgres database
class PostgresConnector(IConnector):
    def __init__(self, db, user, password, host='localhost', port=5432):
        """
        configure connection
        :param db: db-name
        """
        self.db = db
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.connected = False

        self.conn = None
        self.lock = Lock()

        self.__connect()

    def __del__(self):
        if self.conn:
            self.conn.close()

    @retry_log_error()
    def __connect(self):
        self.conn = psycopg2.connect(dbname=self.db, user=self.user,
                                     password=self.password, host=self.host, port=self.port)
        self.cursor = self.conn.cursor()

    @log_error
    def exec(self, request: str, identifiers=None, literals=None, result='all'):
        request = request.replace(symbols.QRDB_IDENTIFIER, '{}')
        request = request.replace(symbols.QRDB_LITERAL, '%s')
        request = sql.SQL(request)
        if identifiers:
            identifiers = [sql.Identifier(x) for x in identifiers]
            request = request.format(*identifiers)

        qrlogging.info('POSTGRES EXECUTE: %s with literals %s', request.as_string(self.cursor), literals)
        with self.lock:
            self.cursor.execute(request, literals)
            data = self.extract_result(result)
        return DBResult(data, result)

    def extract_result(self, result):
        if result == 'all':
            return self.cursor.fetchall()
        elif result == 'one':
            return self.cursor.fetchone()
        elif result is not None:
            qrlogging.warning("unexpected 'result' value: %s" % result)
        return None

    @log_error
    def table_info(self):
        request = '''select table_name, column_name, data_type
                     from information_schema.columns
                     where table_schema = 'public';'''

        data = self.exec(request, result='all')
        info = {}
        for d in data.get_data():
            if not info.get(d[0]):
                info[d[0]] = []
            info[d[0]].append((d[1], d[2]))
        return info

    @log_error
    def commit(self):
        self.conn.commit()



import inject
from data_formatter import *
from IConnector import *
from PostgresConnector import PostgresConnector

def config_DB(binder, connector, data_formatter):
    binder.bind(IDataFormatter, data_formatter())
    binder.bind(IConnector, connector())


def register(format_type, connector_type, *conn_args, **conn_kwargs):
    if connector_type == 'postgres':
        connector = lambda: PostgresConnector(*conn_args, **conn_kwargs)
    else:
        raise Exception('unknown connector type')

    if format_type == 'list':
        data_formatter = ListDataFormatter
    elif format_type in ['dict', None]:
        data_formatter = DictDataFormatter
    else:
        raise Exception('unknown format type')

    inject.clear_and_configure(lambda binder: config_DB(binder, connector, data_formatter))


# default empty configuration. used to avoid import errors on not-configured injections
if not inject.is_configured():
    inject.clear_and_configure(lambda binder: None)


import setuptools

setuptools.setup(
    name="qrookDB",
    version="1.2.0.1",
    author="Kurush",
    author_email="ze17@ya.ru",
    description="tiny ORM for SQL-databases",
    long_description_content_type="text/markdown",
    url="https://gitlab.com/qrook/db",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    long_description='''
# Usage example:
```python
    import DB as db
from data import QRTable
books, books_authors, authors, events = [QRTable()] * 4

DB = db.DB('postgres', 'db_name', 'user_name', 'password', format_type='dict')
op = DB.operators
DB.create_logger(app_name='app', file='app_log.log')
DB.create_data(__name__, in_module=True)

print(DB.books)
print(books, books.id)
data = DB.select(books).where(original_publication_year=2000, language_code='eng').\
    where(id=op.In(470, 490, 485)).all()
print(data[:10])

data = books.select('count(*)').group_by(books.original_publication_year).all()
print(data[:10])

data = DB.select(books, books.id).where('id < 10').order_by(books.id, desc=True).\
    limit(3).offset(2).all()
print('limit & offset', data)

data = books.select(authors.id, books.id)\
    .join(books_authors, op.Eq(books_authors.book_id, books.id))\
    .join(authors, op.Eq(books_authors.author_id, authors.id)).all()
    # .join(books_authors, 'books_authors.book_id = books.id')\
print(data[:10])

data = books.select(books.id).where(id=1).where(bool='or', id=2).limit(10).all()
print('with or:', data)

ok = DB.delete(events, auto_commit=True).where(id=1).exec()
print(ok)

from datetime import datetime
t = datetime.now().time()
d = datetime.now().date()
ok = DB.update(events, auto_commit=True).set(time=t).where(id=6).exec()
print(ok)

#ok = DB.insert(events, events.time, auto_commit=True).values([t]).exec()
#ok = DB.insert(events, events.date, events.time, auto_commit=True).values([d, t]).exec()
query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning('*')
data = query.all()
print(data)

query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning(events.date, events.time)
data = query.all()
print(data)

query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning(['date', 'time'])
data = query.all()
print(data)

query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning('date', 'time')
data = query.all()
print(data)

data = DB.exec('select * from get_book_authors(1) as f(id int, name varchar)').config_fields('id', 'name').all()
print(data)

q = DB.select(events, events.id, events.id).where(id=1)
data = q.all()
print('data:', data, ';\terror:', q.get_error())
```
'''
)



from abc import ABCMeta, abstractmethod, abstractproperty
from symbols import QRDB_IDENTIFIER, QRDB_LITERAL

class IQROperator:
    """
    Abstract class representing sql operator.
    The only purpose is to create the condition string and literals literals for query.
    All operators are supposed to work like this
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def condition(self, short_name=False):
        """
        :param short_name: whether to use a short name or a full one (with table name or not)
        :return: pair: <request_string>, <literals array>
        """


class QROperator(IQROperator):
    def __init__(self, op, literals):
        self._literals = literals
        self._op = op

    @staticmethod
    def _get_name(short_name):
        return QRDB_IDENTIFIER + ' ' if short_name else '%s.%s ' % (QRDB_IDENTIFIER, QRDB_IDENTIFIER)

    def condition(self, short_name=False):
        return self._get_name(short_name) + self._op + QRDB_LITERAL, [self._literals]


class Between(QROperator):
    """
    between operator
    """
    def __init__(self, a, b):
        super().__init__('between', [a, b])

    def condition(self, short_name=False):
        return self._get_name(short_name) + ' between %s and %s' % (QRDB_LITERAL, QRDB_LITERAL), \
               list(self._literals)


class In(QROperator):
    """
    in operator
    """
    def __init__(self, *args):
        """
        :param args: iterable of elements (possible to be casted to a list)
        """
        super().__init__('in', args)

    def condition(self, short_name=False):
        ins = ','.join([QRDB_LITERAL] * len(self._literals))
        return super()._get_name(short_name) + ' in(' + ins + ')', list(self._literals)


class Eq(QROperator):
    """
    equal operator
    todo not very beautiful
    """
    def __init__(self, arg1, arg2=None):
        """
        :param arg1 - literal (or identifier, if arg2 set), right side of '='
        :param arg2: optional. if set, both args are identifiers' names
        examples of 'condition' call:
        Eq(16).condition() -> '{id}={literal}', [2000]
        Eq('a_id', 'b_id') -> '{id}={id}', []
        """
        self.arg1, self.arg2, self.duo = arg1, arg2, arg2 is not None
        super().__init__('=', arg1)

    def has_both_args(self):
        return self.duo

    def condition(self, short_name=False):
        if not self.duo:
            return super().condition(short_name)
        else:
            return self._get_name(short_name) + ' = ' + self._get_name(short_name), []


class GT(QROperator):
    """
    greater operator
    """
    def __init__(self, literals):
        super().__init__('>', literals)


class GE(QROperator):
    """
    greater or equal operator
    """
    def __init__(self, literals):
        super().__init__('>=', literals)


class LT(QROperator):
    """
    less operator
    """
    def __init__(self, literals):
        super().__init__('<', literals)


class LE(QROperator):
    """
    less or equal operator
    """
    def __init__(self, literals):
        super().__init__('<=', literals)


class NE(QROperator):
    """
    not equal operator
    """
    def __init__(self, literals):
        super().__init__('<>', literals)


class Like(QROperator):
    """
    like operator
    """
    def __init__(self, literals):
        super().__init__(' like ', literals)
    # todo class like - update condition() method


import logging

_log_format = "%(asctime)s [%(levelname)s]: [%(app)s, %(name)s] - " \
              "(%(filename)s).%(funcName)s(%(lineno)d) - %(message)s"

logger = None

def info(msg, *args):
    if logger:
        logger.info(msg, *args)

def warning(msg, *args):
    if logger:
        logger.warning(msg, *args)

def error(msg, *args):
    if logger:
        logger.error(msg, *args)

def exception(msg, *args):
    if logger:
        logger.exception(msg, *args)


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""
    grey = "\033[30m"
    yellow = "\033[33m"
    red = "\033[31m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    FORMATS = {
        logging.DEBUG: grey + _log_format + reset,
        logging.INFO: grey + _log_format + reset,
        logging.WARNING: yellow + _log_format + reset,
        logging.ERROR: red + _log_format + reset,
        logging.CRITICAL: bold_red + _log_format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_file_handler(filename, level):
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(logging.Formatter(_log_format))
    return file_handler


def get_stream_handler(level):
    sh = logging.StreamHandler()
    sh.setLevel(level)
    sh.setFormatter(logging.Formatter(_log_format))
    sh.setFormatter(CustomFormatter())
    return sh


def create_logger(logger_name='default', app_name='app', level="INFO",
                  file: str = None, file_level="INFO"):
    logger = logging.getLogger(logger_name)
    logger.addHandler(get_stream_handler(level))
    if file:
        logger.addHandler(get_file_handler(file, file_level))

    logger = logging.LoggerAdapter(logger, {'app': app_name})
    logger.setLevel('INFO')
    return logger


import qrlogging
import time


def log_error(f):
    def wrapper(*args, **kwargs):
        try:
            s = f(*args, **kwargs)
            return s
        except Exception as e:
            qrlogging.exception(e)

    return wrapper


def log_error_default(default=None):
    def decorator(f):
        def wrapper(*args, **kwargs):
            try:
                s = f(*args, **kwargs)
                return s
            except Exception as e:
                qrlogging.exception(e)
                return default

        return wrapper

    return decorator


def log_error_default_self(f):
    def wrapper(*args, **kwargs):
        try:
            s = f(*args, **kwargs)
            return s
        except Exception as e:
            qrlogging.exception(e)
            return args[0]

    return wrapper


def retry_log_error(retry_delay=5):
    def decorator(f):
        def wrapper(*args, **kwargs):
            while 1:
                try:
                    s = f(*args, **kwargs)
                    return s
                except Exception as e:
                    qrlogging.warning(str(e) + '; retrying in %s s...' % retry_delay)
                    time.sleep(retry_delay)

        return wrapper

    return decorator


def log_class(decorator, exceptions=None):
    if exceptions is None:
        exceptions = []

    def logger(cls):
        for o in dir(cls):
            if o.startswith('__') or o in exceptions:
                continue
            a = getattr(cls, o)
            if hasattr(a, '__call__'):
                decorated_a = decorator(a)
                setattr(cls, o, decorated_a)
        return cls

    return logger


def exc_no_db(exceptions=None):
    if exceptions is None:
        exceptions = []

    def exc(f):
        def decorator(self, *args, **kwargs):
            if self._DB is None:
                raise Exception('no DB instance set for object to make queries')
            return f(self, *args, **kwargs)

        return decorator

    def logger(cls):
        for o in dir(cls):
            if o.startswith('__') or o in exceptions: continue
            a = getattr(cls, o)
            if hasattr(a, '__call__'):
                decorated_a = exc(a)
                setattr(cls, o, decorated_a)
        return cls

    return logger
